{-# LANGUAGE DataKinds #-}
{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE RecordWildCards #-}
{-# LANGUAGE ExplicitNamespaces #-}
{-# LANGUAGE NoStarIsType #-}
{-# LANGUAGE AllowAmbiguousTypes #-}
{-# LANGUAGE GADTs #-}
{-# LANGUAGE KindSignatures #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE PolyKinds #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE UndecidableSuperClasses #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE TupleSections #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE Strict #-}
{-# OPTIONS_GHC -fplugin GHC.TypeLits.KnownNat.Solver #-}
{-# OPTIONS_GHC -fno-full-laziness #-}

module Synthesizer.Train (module Synthesizer.Train) where

import           System.Random                 (StdGen, mkStdGen, setStdGen)
import           System.Timeout                (timeout)
import           System.Directory              (createDirectoryIfMissing, doesFileExist, removeFile)
import           System.CPUTime
import           System.ProgressBar
import qualified Data.Text.Lazy as T
import           Data.Text.Internal.Lazy (Text)
import qualified Data.Text as Text
import           Data.Maybe                    (fromMaybe)
import           Data.Set                      (Set, empty, insert)
import qualified Data.Set as Set
import           Data.Bifunctor                (first, second)
import qualified Data.ByteString               as BS
import qualified Data.ByteString.Internal      as BS
import qualified Data.ByteString.Lazy.Internal as BL
import           Data.HashMap.Lazy             (HashMap, (!), elems, keys, size, mapWithKey, filterWithKey, fromListWith, singleton, findWithDefault)
import qualified Data.Csv as Csv
import           Data.Text.Prettyprint.Doc (pretty)
import           Text.Printf
import           Foreign.Marshal.Utils         (fromBool)
import           Control.Monad                 (join, replicateM, forM, void, when)
import           Control.Monad.Trans.Loop
import           Control.Monad.Trans.Class
import           Control.Monad.Logger
import           Language.Haskell.Exts.Syntax  ( Exp (..) )
import           Prelude                        hiding (abs)
import           Util                          (fstOf3)
import           Language.Haskell.Interpreter  ( Interpreter, liftIO, lift )
import           GHC.Exts
import           GHC.Generics                  (Generic)
import           GHC.TypeNats                  (KnownNat, Nat, CmpNat, type (*), type (-))
import qualified Torch.Functional.Internal     as I
import qualified Torch.DType                   as D
import qualified Torch.Tensor                  as D
import qualified Torch.Device                  as D
import qualified Torch.TensorFactories         as D
import qualified Torch.TensorOptions           as D
import qualified Torch.Optim                   as D
import qualified Torch.Serialize               as D
import qualified Torch.Autograd                as D
import qualified Torch.Functional              as F
import qualified Torch.NN                      as A
import           Torch.Tensorboard
import           Torch.Typed.NN.Recurrent.LSTM
import           Torch.Typed.Aux
import           Torch.Typed.Tensor
import           Torch.Typed.NN
import           Torch.Typed.Parameter
import qualified Torch.Typed.Parameter
import           Torch.Typed.Factories
import           Torch.Typed.Optim
import           Torch.Typed.Functional
import           Torch.Typed.Autograd
import           Torch.Typed.Serialize
import qualified Torch.Distributions.Distribution as Distribution
import qualified Torch.Distributions.Categorical as Categorical

import           Synthesis.Orphanage ()
import           Synthesis.Data hiding (GridSearchConfig(..), EvolutionaryConfig(..))
import           Synthesis.Configs
import           Synthesis.Utility
import           Synthesis.Ast
import           Synthesis.Generation
import           Synthesis.FindHoles
import           Synthesis.Hint
import           Synthesis.Types
import           Synthesizer.Utility
import           Synthesizer.Encoder
import           Synthesizer.R3NN
import           Synthesizer.Synthesizer
import           Synthesizer.Params

-- -- | deterministically pick the most likely expansion to fill a hole in a PPT
-- -- | deprecated, not in use
-- -- TODO: replace this whole block with argmaxAll (argmax_t) returning non-flat index (like np.unravel_index)...
-- -- (hole_idx, rule_idx) :: (Int, Int) = unravelIdx . argmaxAll $ hole_expansion_probs
-- argmaxExpansion :: forall num_holes rules device . Tensor device 'D.Float '[num_holes, rules] -> (Int, Int)
-- argmaxExpansion hole_expansion_probs = (hole_idx, rule_idx) where
--     (hole_dim, rule_dim) :: (Int, Int) = (0, 1)
--     rule_idx_by_hole :: Tensor device 'D.Int64 '[num_holes] =
--             asUntyped (F.argmax (F.Dim rule_dim) F.RemoveDim) hole_expansion_probs
--     num_holes_ :: Int = shape' rule_idx_by_hole !! 0
--     best_prob_by_hole :: Tensor device 'D.Float '[num_holes] =
--             UnsafeMkTensor . D.reshape [num_holes_] $ I.gather  -- F.squeeze 1
--                 (toDynamic hole_expansion_probs)
--                 rule_dim
--                 (toDynamic $ unsqueeze @1 rule_idx_by_hole)
--                 False
--     hole_idx :: Int = D.asValue $ F.argmax (F.Dim 0) F.RemoveDim $ toDynamic best_prob_by_hole
--     rule_idx :: Int = D.asValue $ D.select 0 hole_idx $ toDynamic rule_idx_by_hole

-- | fill a non-terminal leaf node in a PPT given hole/rule expansion probabilities
predictHole :: forall num_holes rules device
                . (StandardFloatingPointDTypeValidation device 'D.Float)
                => Bool
                -> [(String, Expr)]
                -> Expr
                -> Set String
                -> Tensor device 'D.Float '[num_holes, rules]
                -> IO (Expr, Set String)
predictHole randomHole variants ppt used hole_expansion_probs = do
    debug_ "predictHole"
    let (holes_dim, _rules_dim) :: (Int, Int) = (0, 1)
    [hole_idx, rule_idx] :: [Int] <- if randomHole then
            sampleIdxs . softmaxAll . toDynamic $ hole_expansion_probs
        else do
            let hole_idx :: Int = 0
            let holeScores :: Tensor device 'D.Float '[rules] = UnsafeMkTensor $ D.select holes_dim hole_idx $ toDynamic hole_expansion_probs
            let holeProbs  :: Tensor device 'D.Float '[rules] = softmax @0 holeScores
            [rule_idx] :: [Int] <- D.asValue <$> Distribution.sample (Categorical.fromProbs . toDynamic $ holeProbs) [1]
            return [hole_idx, rule_idx]
    -- order of rules: comes from `rule_emb`, which is just randomly assigned,
    -- so we can just arbitrarily associate this with any deterministic order e.g. that of `variants`
    debug_ $ "rule_idx: " <> show rule_idx
    let (_rule_str, rule_expr) :: (String, Expr) = variants !! rule_idx
    let block_name :: String = pp . head . fnAppNodes $ rule_expr
    debug_ $ "block_name: " <> show block_name
    let used' :: Set String = insert block_name used
    -- grab hole lenses by `findHolesExpr` and ensure `forwardPass` follows the same order to make these match.
    let (_hole_getter, hole_setter) :: (Expr -> Expr, Expr -> Expr -> Expr) =
            findHolesExpr ppt !! hole_idx
    let ppt' :: Expr = hole_setter ppt rule_expr
    debug_ . show $ (hole_idx, rule_idx, pp ppt')
    return (ppt', used')

-- | fill a random non-terminal leaf node as per `task_fn`
superviseHole :: forall device
                . ( KnownDevice device
                  , RandDTypeIsValid device 'D.Int64
                  )
                => Bool
                -> HashMap String Expr
                -> Int
                -> Expr
                -> Expr
                -> IO (Expr, String)
superviseHole randomHole variantMap num_holes task_fn ppt = do
    debug_ "superviseHole"
    hole_idx :: Int <- if randomHole then do
            -- randomly pick a hole -- probably a good way to uniformly cover all scenarios...?
            -- technically holes closer to the root may go thru more rounds tho, creating more opportunities to randomly get picked
            hole_idx' :: Tensor device 'D.Int64 '[] <- randint 0 num_holes
            return $ toInt hole_idx'
        -- otherwise deterministically use the first hole
        else pure 0
    debug_ $ "hole_idx: " <> show hole_idx
    let (hole_getter, hole_setter) :: (Expr -> Expr, Expr -> Expr -> Expr) =
            findHolesExpr ppt !! hole_idx
    -- identify what block is used at this hole in the task function we are modeling,
    -- then just mimic that so we can proceed to supervising the next prediction step.
    let rule_expr :: Expr = safeIndexHM variantMap . nodeRule . hole_getter $ task_fn
    debug_ $ "rule_expr: " <> pp rule_expr
    let block_name :: String = pp . head . fnAppNodes $ rule_expr
    let ppt' :: Expr = hole_setter ppt rule_expr
    debug_ $ "ppt': " <> pp ppt'
    return (ppt', block_name)

-- | supervise with task program to calculate the loss of the predicted hole/rule expansion probabilities for this PPT
fillHoleTrain :: forall num_holes rules device
                . ( KnownDevice device
                  , RandDTypeIsValid device 'D.Int64
                  )
                => Bool
                -> HashMap String Expr
                -> HashMap String Int
                -> Expr
                -> Expr
                -> Tensor device 'D.Float '[num_holes, rules]
                -> IO (Expr, String, Tensor device 'D.Float '[num_holes])
fillHoleTrain randomHole variantMap ruleIdxs task_fn ppt hole_expansion_probs = do
    debug_ "fillHoleTrain"
    let (_hole_dim, rule_dim) :: (Int, Int) = (0, 1)
    let [num_holes, _rules] :: [Int] = shape' hole_expansion_probs
    debug_ $ "num_holes: " <> show num_holes
    (ppt', block_name) :: (Expr, String) <- superviseHole @device randomHole variantMap num_holes task_fn ppt
    debug_ $ "ppt': " <> pp ppt'
    -- iterate over holes to get their intended expansion 'probabilities', used in calculating the loss
    let gold_rule_probs :: Tensor device 'D.Float '[num_holes] = UnsafeMkTensor . D.toDevice (deviceVal @device) . D.asTensor $ getGold . fst <$> findHolesExpr ppt
            where getGold = \gtr -> safeIndexHM ruleIdxs . nodeRule . gtr $ task_fn
    return (ppt', block_name, gold_rule_probs)

-- | zero out any predicted probabilities of non-compiling programs
maskUncompiling :: forall device num_holes rules
          . ( KnownDevice device
            )
          => Tensor device 'D.Float '[num_holes, rules]
          -> Expr
          -> Set String
          -> HashMap String Expr
          -> [(String, Expr)]
          -> Interpreter (Tensor device 'D.Float '[num_holes, rules])
maskUncompiling predicted ppt used dsl variants = do
    --  :: Tensor device 'D.Float '[num_holes, rules]
    mask <-
        fmap (Torch.Typed.Tensor.toDType @'D.Float . UnsafeMkTensor . D.asTensor) $
        (\(hole_getter, hole_setter) -> mapM (
                \(_rule_str, rule_expr) -> let
                        block_name :: String = pp . head . fnAppNodes $ rule_expr
                        used' :: Set String = insert block_name used
                        defs :: HashMap String Expr = pickKeysSafe (Set.toList used') dsl
                    in fitExpr' . iff (not . null $ defs) (letIn defs) . hole_setter ppt $ rule_expr
            ) variants)
        `mapM` findHolesExpr ppt
    return . Torch.Typed.Tensor.toDevice @device . asUntyped (F.mul $ toDynamic mask) $ predicted

-- | calculate the loss by comparing the predicted expansions to the intended programs
calcLoss :: forall rules ruleFeats device shape synthesizer num_holes
          . ( KnownDevice device
            , MatMulDTypeIsValid device 'D.Float
            , SumDTypeIsValid device 'D.Float
            , BasicArithmeticDTypeIsValid device 'D.Float
            , RandDTypeIsValid device 'D.Int64
            , KnownNat rules
            , KnownNat ruleFeats
            , Synthesizer device shape rules ruleFeats synthesizer
            )
          => Bool
          -> HashMap String Expr
          -> Expr
          -> Tp
          -> HashMap String Int
          -> synthesizer
          -> Tensor device 'D.Float shape
          -> HashMap String Expr
          -> HashMap String Int
          -> HashMap String Int
          -> Int
          -> Bool
          -> [(String, Expr)]
          -> Set String
          -> Tensor device 'D.Float '[rules, ruleFeats]
          -> Interpreter (Tensor device 'D.Float '[])
calcLoss randomHole dsl task_fn taskType symbolIdxs model io_feats variantMap ruleIdxs variant_sizes max_holes maskBad variants used rule_tp_emb = do
    debug "calcLoss"
    let (_hole_dim, rule_dim) :: (Int, Int) = (0, 1)
    (_program, _used, golds, predictions, _filled) :: (Expr, Set String, [D.Tensor], [D.Tensor], Int) <- let
            fill = \(ppt, used_, golds, predictions, filled) -> do
                    --  :: Tensor device 'D.Float '[num_holes, rules]
                    let predicted = predict @device @shape @rules @ruleFeats @synthesizer model symbolIdxs ppt rule_tp_emb io_feats
                    -- debug $ "predicted: " <> show (shape' predicted)
                    predicted' <- if not maskBad then pure predicted else maskUncompiling predicted (unLet ppt) used_ dsl variants
                    -- debug $ "predicted': " <> show (shape' predicted')
                    (ppt', block_name, gold) <- liftIO $ fillHoleTrain randomHole variantMap ruleIdxs task_fn ppt predicted'
                    debug $ "ppt': " <> pp ppt'
                    let used'' :: Set String = insert block_name used_
                    return (ppt', used'', (:) (toDynamic gold) $! golds, (:) (toDynamic predicted') $! predictions, filled + 1)
            in while_ (\(expr, _, _, _, filled) -> hasHoles expr && filled < max_holes) fill (letIn dsl (skeleton taskType), used, [], [], 0 :: Int)
    let gold_rule_probs :: D.Tensor = F.cat (F.Dim 0) golds
    -- debug $ "gold_rule_probs: " <> show (D.shape gold_rule_probs)
    let hole_expansion_probs :: D.Tensor = F.cat (F.Dim 0) predictions
    -- debug $ "hole_expansion_probs: " <> show (D.shape hole_expansion_probs)
    let loss :: Tensor device 'D.Float '[] = patchLoss @device @shape @rules @ruleFeats model variant_sizes $ UnsafeMkTensor $ crossEntropy gold_rule_probs rule_dim hole_expansion_probs
    -- debug $ "loss: " <> show (shape' loss)
    return loss

-- | pre-calculate DSL stuff
prep_dsl :: TaskFnDataset -> PreppedDSL
prep_dsl TaskFnDataset{..} =
    PreppedDSL variants variant_sizes symbolIdxs ruleIdxs variantMap max_holes dsl' fnUsed
    where
    variants :: [(String, Expr)] = (\(_k, v) -> (nodeRule v, v)) <$> exprBlocks
    variant_sizes :: HashMap String Int = fromList $ variantInt . snd <$> variants
    symbolIdxs :: HashMap String Int = indexList $ "undefined" : keys dsl
    ruleIdxs :: HashMap String Int = indexList $ fst <$> variants
    variantMap :: HashMap String Expr = fromList variants
    -- for synthesized programs, we apply the same maximum number of holes as used to generate this dataset. this allows our synthesizer enough power to construct the desired programs, while disallowing more complex programs than those of the maximum generated complexity. this choice is arbitrary; yet to think of a nicer solution.
    max_holes :: Int = maxHoles generationCfg
    -- DSL without entries equal to their key, for constructing let-in expressions.
    -- without this, blocks identical to their keys are seen as recursive, causing non-termination
    dsl' = filterWithKey (\k v -> k /= pp v) dsl
    fnUsed :: HashMap Expr (Set String) = fromKeys (Set.fromList . fmap pp . subExprs) . keys $ fnTypes

-- | train a NSPS model and return results
train :: forall device rules shape ruleFeats synthesizer
        . ( KnownDevice device
          , RandDTypeIsValid device 'D.Float
          , MatMulDTypeIsValid device 'D.Float
          , SumDTypeIsValid device 'D.Float
          , BasicArithmeticDTypeIsValid device 'D.Float
          , RandDTypeIsValid device 'D.Int64
          , StandardFloatingPointDTypeValidation device 'D.Float
          , KnownNat rules
          , KnownNat ruleFeats
          , KnownShape shape
          , Synthesizer device shape rules ruleFeats synthesizer
          , KnownNat (FromMaybe 0 (ExtractDim BatchDim shape))
          )
          => SynthesizerConfig
          -> TaskFnDataset
          -> synthesizer
          -> IO [EvalResult]
train synthesizerConfig taskFnDataset init_model = do
    debug_ "train"
    let SynthesizerConfig{..} = synthesizerConfig
    let TaskFnDataset{..} = taskFnDataset

    -- GENERAL

    let [train_set, validation_set, test_set] :: [[(Expr, (Tp, Tp))]] = lists2pairs <$> (if cheat then replicate 3 . fstOf3 else untuple3) datasets
    let [train_blocks, validation_blocks, test_blocks] :: [HashMap String Int] = tally . (=<<) (pp <.> subExprs . fst) . lists2pairs <.> untuple3 $ datasets
    notice_ $ "train_blocks:\n"      <> show train_blocks
    notice_ $ "validation_blocks:\n" <> show validation_blocks
    notice_ $ "test_blocks:\n"       <> show test_blocks
    let n :: Int = length train_set
    let stdGen :: StdGen = mkStdGen seed
    setStdGen stdGen
    let init_lr :: Tensor device 'D.Float '[] = UnsafeMkTensor . D.asTensor $ learningRate
    let runName :: String = ppCfg synthesizerConfig
    let modelFolder = resultFolder <> "/" <> runName
    createDirectoryIfMissing True resultFolder
    createDirectoryIfMissing True modelFolder

    let resultPath = resultFolder <> "/" <> runName <> ".csv"
    when (null savedModelPath) $ do
        exists :: Bool <- doesFileExist resultPath
        if exists
            then error $ "path " <> resultPath <> " exists!"
            else BS.writeFile resultPath $ BS.packChars $ BL.unpackChars $ Csv.encodeByName evalResultHeader ([] :: [EvalResult])

    let prepped_dsl = prep_dsl taskFnDataset
    let PreppedDSL{..} = prepped_dsl
    debug_ $ "variants: " <> pp_ variants
    debug_ $ "variantMap: " <> pp_ variantMap
    debug_ $ "symbolIdxs: " <> show symbolIdxs
    debug_ $ "ruleIdxs: " <> show ruleIdxs

    sw <- Torch.Tensorboard.summaryWriter (Text.pack $ resultFolder <> "/tensorboard") $ Text.pack runName
    Torch.Tensorboard.setStep sw initialEpoch

    -- MODELS
    let init_optim :: D.Adam = d_mkAdam 0 0.9 0.999 $ A.flattenParameters init_model
    let init_state = (stdGen, init_model, init_optim, False, [], init_lr, 0.0, initialEpoch)

    (_, model, _, _, eval_results, _, _, _) <- iterateLoopT init_state $ \ !state@(gen, model_, optim_, earlyStop, eval_results, lr, prev_acc, epoch) -> if earlyStop || epoch > numEpochs then exitWith state else do

        (loss_train, model', optim', gen'', epochSeconds) <- lift $ if epoch == 0 then do
            loss <- evaluateLoss @device @rules @shape @ruleFeats taskFnDataset prepped_dsl maskBad randomHole model_ train_set
            return (loss, model_, optim_, gen, 0.0/0.0)
            else do
                let (train_set', gen') = fisherYates gen train_set    -- shuffle
                pb <- newProgressBar pgStyle 1 (Progress 0 n (T.pack $ "epoch " <> show epoch <> " task-fns"))
                start <- getCPUTime
                -- TRAIN LOOP
                (loss_train, model', optim', gen'', _) :: (Float, synthesizer, D.Adam, StdGen, Int) <- iterateLoopT (0.0, model_, optim_, gen', 0) $ \ !state@(train_loss, model, optim, gen_, task_fn_id) -> if task_fn_id >= n then exitWith state else do
                        let (task_fn, tpInstPair) :: (Expr, (Tp, Tp)) = train_set' !! task_fn_id
                        -- lift . debug_ $ "task_fn: \n" <> pp task_fn
                        let taskType :: Tp = safeIndexHM fnTypes task_fn
                        lift . debug_ $ "taskType: " <> pp taskType
                        let used :: Set String = safeIndexHM fnUsed task_fn
                        let (target_tp_io_pairs, gen') :: (HashMap (Tp, Tp) [(Expr, Either String Expr)], StdGen) =
                                first (singleton tpInstPair) . fixSize (natValI @R3nnBatch) gen_ $ safeIndexHM (safeIndexHM fnTypeIOs task_fn) tpInstPair
                        lift . debug_ $ "target_tp_io_pairs: " <> pp_ target_tp_io_pairs
                        let rule_tp_emb :: Tensor device 'D.Float '[rules, ruleFeats] =
                                rule_encode @device @shape @rules @ruleFeats model variantTypes
                        -- lift . debug_ $ "rule_tp_emb: " <> show (shape' rule_tp_emb)
                        --  :: Tensor device 'D.Float '[n'1, t * (2 * Dirs * h)]
                        -- sampled_feats :: Tensor device 'D.Float '[R3nnBatch, t * (2 * Dirs * h)]
                        let io_feats :: Tensor device 'D.Float shape = encode @device @shape @rules @ruleFeats model target_tp_io_pairs
                        -- lift . debug $ "io_feats: " <> show (shape' io_feats)
                        loss :: Tensor device 'D.Float '[] <- lift . interpretUnsafe $ calcLoss @rules @ruleFeats randomHole dsl' task_fn taskType symbolIdxs model io_feats variantMap ruleIdxs variant_sizes max_holes maskBad variants used rule_tp_emb
                        -- lift . debug_ $ "loss: " <> show (shape' loss)
                        -- TODO: do once for each mini-batch / fn?
                        -- (newParam, optim') <- D.runStep model optim (toDynamic loss) $ toDynamic lr
                        let params = A.flattenParameters model
                        -- regularize (weight decay) and clip gradients
                        --  . decayWeights' regularization params
                        let gradients = clipGradients' clip $ D.grad' (toDynamic loss) params
                        (newParam, optim') <- lift $ doStep @device @shape @rules @ruleFeats model optim lr gradients
                        let model' :: synthesizer = A.replaceParameters model newParam
                        -- aggregating over task fns, which in turn had separately aggregated over any holes encountered across the different synthesis steps (so multiple times for a hole encountered across various PPTs along the way). this is fair, right?
                        let train_loss' :: Float = train_loss + toFloat loss / fromIntegral n
                        lift $ incProgress pb 1
                        return (train_loss', model', optim', gen', task_fn_id + 1)

                debug_ $ "finished epoch training"
                end <- getCPUTime
                let epochSeconds :: Double = (fromIntegral (end - start)) / (10^12)
                return (loss_train, model', optim', gen'', epochSeconds)

        lift $ Torch.Tensorboard.addScalar sw "epochSeconds" epochSeconds
        lift $ Torch.Tensorboard.addScalar sw "loss_train" loss_train
        let modelPath = modelFolder <> printf "/%04d.pt" epoch
        lift $ D.save (D.toDependent <$> A.flattenParameters model') modelPath

        -- EVAL
        (earlyStop, eval_results', gen''') <- lift $ whenOrM (False, eval_results, gen'') (mod epoch evalFreq == 0) $ do
            debug_ "evaluating"

            (acc_valid, loss_valid) <- evaluate @device @rules @shape @ruleFeats taskFnDataset prepped_dsl bestOf maskBad randomHole model' validation_set

            say_ $ printf
                "Epoch: %03d. Train loss: %.4f. Validation loss: %.4f. Validation accuracy: %.4f.\n"
                epoch
                loss_train
                loss_valid
                acc_valid

            let eval_result = EvalResult epoch epochSeconds loss_train loss_valid acc_valid
            -- write result to csv
            BS.appendFile resultPath . BS.packChars $ BL.unpackChars $ Csv.encodeByNameWith (Csv.defaultEncodeOptions { Csv.encIncludeHeader = False }) evalResultHeader [eval_result]

            Torch.Tensorboard.addScalar sw "loss_valid" loss_valid
            Torch.Tensorboard.addScalar sw "acc_valid" acc_valid

            let eval_results' = (:) eval_result $! eval_results
            let earlyStop :: Bool = whenOr False (length eval_results' >= 2 * checkWindow) $ let
                    losses  :: [Float] = lossValid <$> eval_results'
                    losses' :: [Float] = take (2 * checkWindow) losses
                    (current_losses, prev_losses) = splitAt checkWindow losses'
                    current :: D.Tensor = F.mean . D.asTensor $ current_losses
                    prev    :: D.Tensor = F.mean . D.asTensor $ prev_losses
                    earlyStop :: Bool = D.asValue $ F.sub current prev `I.gtScalar` convergenceThreshold
                    in earlyStop
            when earlyStop $ do
                let msg = "validation loss has converged, stopping early!"
                info_ msg
                Torch.Tensorboard.writeEvent sw . Torch.Tensorboard.EventMessage LevelInfo $ Text.pack msg

            return $ (earlyStop, eval_results', gen'')

        let acc_valid :: Float = if (null eval_results') then prev_acc else accValid $ head eval_results'
        -- decay the learning rate if accuracy decreases
        lr' :: Tensor device 'D.Float '[] <- lift $ case (acc_valid < prev_acc) of
            True -> do
                let msg = "accuracy decreased, decaying learning rate!"
                info_ $ msg
                Torch.Tensorboard.writeEvent sw . Torch.Tensorboard.EventMessage LevelInfo $ Text.pack msg
                return . divScalar learningDecay $ lr
            False -> pure lr

        lift $ Torch.Tensorboard.nextStep sw
        return (gen''', model', optim', earlyStop, eval_results', lr', acc_valid, epoch + 1)

    say_ $ "data written to " <> resultPath

    return $ reverse eval_results

evaluate :: forall device rules shape ruleFeats synthesizer
          . ( KnownDevice device
            , RandDTypeIsValid device 'D.Float
            , MatMulDTypeIsValid device 'D.Float
            , SumDTypeIsValid device 'D.Float
            , BasicArithmeticDTypeIsValid device 'D.Float
            , RandDTypeIsValid device 'D.Int64
            , StandardFloatingPointDTypeValidation device 'D.Float
            , KnownNat rules
            , KnownNat ruleFeats
            , KnownShape shape
            , Synthesizer device shape rules ruleFeats synthesizer
            , KnownNat (FromMaybe 0 (ExtractDim BatchDim shape))
            )
         => TaskFnDataset
         -> PreppedDSL
         -> Int
         -> Bool
         -> Bool
         -> synthesizer
         -> [(Expr, (Tp, Tp))]
         -> IO (Float, Float)
evaluate taskFnDataset preppedDSL bestOf maskBad randomHole model dataset = do
    loss <- evaluateLoss @device @rules @shape @ruleFeats taskFnDataset preppedDSL maskBad randomHole model dataset
    acc <-  evaluateAcc  @device @rules @shape @ruleFeats taskFnDataset preppedDSL maskBad randomHole model dataset bestOf
    return (acc, loss)

evaluateLoss :: forall device rules shape ruleFeats synthesizer
          . ( KnownDevice device
            , RandDTypeIsValid device 'D.Float
            , MatMulDTypeIsValid device 'D.Float
            , SumDTypeIsValid device 'D.Float
            , BasicArithmeticDTypeIsValid device 'D.Float
            , RandDTypeIsValid device 'D.Int64
            , StandardFloatingPointDTypeValidation device 'D.Float
            , KnownNat rules
            , KnownNat ruleFeats
            , KnownShape shape
            , Synthesizer device shape rules ruleFeats synthesizer
            , KnownNat (FromMaybe 0 (ExtractDim BatchDim shape))
            )
         => TaskFnDataset
         -> PreppedDSL
         -> Bool
         -> Bool
         -> synthesizer
         -> [(Expr, (Tp, Tp))]
         -> IO Float
evaluateLoss TaskFnDataset{..} PreppedDSL{..} maskBad randomHole model dataset = do
    debug_ "evaluateLoss"
    let rule_tp_emb :: Tensor device 'D.Float '[rules, ruleFeats] =
                rule_encode @device @shape @rules @ruleFeats model variantTypes

    let n :: Int = length dataset
    pb <- newProgressBar pgStyle 1 $ Progress 0 n ("eval-loss" :: Text)
    (loss, _) :: (Float, Int) <- iterateLoopT (0.0, 0) $ \ !state@(loss, task_fn_id) -> if task_fn_id >= n then exitWith state else do
            let (task_fn, tpInstPair) :: (Expr, (Tp, Tp)) = dataset !! task_fn_id
            -- lift . debug $ "task_fn: \n" <> pp task_fn
            let taskType :: Tp = safeIndexHM fnTypes task_fn
            lift . debug_ $ "taskType: " <> pp taskType
            let used :: Set String = safeIndexHM fnUsed task_fn
            let target_tp_io_pairs :: HashMap (Tp, Tp) [(Expr, Either String Expr)] =
                    singleton tpInstPair $ safeIndexHM (safeIndexHM fnTypeIOs task_fn) tpInstPair

            let io_feats :: Tensor device 'D.Float shape = encode @device @shape @rules @ruleFeats model target_tp_io_pairs
            loss_ :: Tensor device 'D.Float '[] <- lift . interpretUnsafe $ calcLoss @rules @ruleFeats randomHole dsl' task_fn taskType symbolIdxs model io_feats variantMap ruleIdxs variant_sizes max_holes maskBad variants used rule_tp_emb

            let loss' :: Float = loss + toFloat loss_      / fromIntegral n
            lift $ incProgress pb 1
            return (loss', task_fn_id + 1)

    return loss

evaluateAcc :: forall device rules shape ruleFeats synthesizer num_holes
          . ( KnownDevice device
            , RandDTypeIsValid device 'D.Float
            , MatMulDTypeIsValid device 'D.Float
            , SumDTypeIsValid device 'D.Float
            , BasicArithmeticDTypeIsValid device 'D.Float
            , RandDTypeIsValid device 'D.Int64
            , StandardFloatingPointDTypeValidation device 'D.Float
            , KnownNat rules
            , KnownNat ruleFeats
            , KnownShape shape
            , Synthesizer device shape rules ruleFeats synthesizer
            , KnownNat (FromMaybe 0 (ExtractDim BatchDim shape))
            )
         => TaskFnDataset
         -> PreppedDSL
         -> Bool
         -> Bool
         -> synthesizer
         -> [(Expr, (Tp, Tp))]
         -> Int
         -> IO Float
evaluateAcc TaskFnDataset{..} PreppedDSL{..} maskBad randomHole model dataset bestOf = do
    -- debug_ "evaluateAcc"
    let rule_tp_emb :: Tensor device 'D.Float '[rules, ruleFeats] =
                rule_encode @device @shape @rules @ruleFeats model variantTypes

    let fldr = resultFolderDef  -- TODO: use chosen resultFolder
    outcomes :: [Bool] <- trackMapMList fldr "eval-acc" dataset $ \tpl -> do
            let (task_fn, tpInstPair) :: (Expr, (Tp, Tp)) = tpl
            -- debug_ $ "task_fn: \n" <> pp task_fn
            let taskType :: Tp = safeIndexHM fnTypes task_fn
            -- debug_ $ "taskType: " <> pp taskType
            let target_tp_io_pairs :: HashMap (Tp, Tp) [(Expr, Either String Expr)] =
                    singleton tpInstPair $ safeIndexHM (safeIndexHM fnTypeIOs task_fn) tpInstPair
            let type_ins :: HashMap (Tp, Tp) [Expr] = fmap fst <$> target_tp_io_pairs
            -- debug_ $ "type_ins: " <> pp_ type_ins
            let target_outputs :: [Either String Expr] =
                    fmap snd $ join . elems $ target_tp_io_pairs

            let io_feats :: Tensor device 'D.Float shape = encode @device @shape @rules @ruleFeats model target_tp_io_pairs

            -- sample for best of 100 predictions
            -- TODO: dedupe samples before eval to save evals?
            -- TODO: consider A* / branch-and-bound / beam search instead
            -- pb <- newProgressBar pgStyle 1 $ Progress 0 bestOf ("eval-samples" :: Text)
            iterateLoopT (False, 0) $ \ !state@(works, try_num) -> if works || try_num >= bestOf then exitWith works else do 
                (program, used, _filled) :: (Expr, Set String, Int) <- let
                        --  :: (Int, Expr) -> IO (Int, Expr)
                        fill = \(ppt, used, filled) -> lift $ do
                                --  :: Tensor device 'D.Float '[num_holes, rules]
                                let predicted = predict @device @shape @rules @ruleFeats @synthesizer model symbolIdxs ppt rule_tp_emb io_feats
                                -- debug_ $ "predicted: " <> show predicted
                                predicted' <- if not maskBad then pure predicted else interpretUnsafe $ maskUncompiling predicted ppt used dsl' variants
                                (ppt', used') <- predictHole randomHole variants ppt used predicted'
                                return (ppt', used', filled + 1)
                        in while_ (\(ppt, used, filled) -> hasHoles ppt && filled < max_holes) fill (skeleton taskType, empty, 0 :: Int)
                -- lift . debug_ $ pp program
                ok :: Bool <- if hasHoles program then pure False else lift $ do
                    let defs :: HashMap String Expr = pickKeysSafe (Set.toList used) dsl'
                    let program' :: Expr = if null defs then program else letIn defs program

                    -- debug_ $ "type_ins: " <> pp_ type_ins
                    prediction_type_ios :: HashMap (Tp, Tp) [(Expr, Either String Expr)] <- let
                            compileInput :: (Tp, Tp) -> [Expr] -> Interpreter [(Expr, Either String Expr)] = \ tp_instantiation ins -> let
                                    n :: Int = length $ unTuple' $ ins !! 0
                                    -- crash_on_error=False is slower but lets me check if it compiles.
                                    in fnIoPairs False n program' tp_instantiation $ list ins
                            in interpretUnsafe . sequence $ compileInput `mapWithKey` type_ins
                    -- debug_ $ "prediction_type_ios: " <> pp_ prediction_type_ios
                    let prediction_io_pairs :: [(Expr, Either String Expr)] =
                            join . elems $ prediction_type_ios
                    let outputs_match :: Bool = case length target_outputs == length prediction_io_pairs of
                            False -> False
                            True -> let
                                    prediction_outputs :: [Either String Expr] = snd <$> prediction_io_pairs
                                    output_matches :: [Bool] = uncurry (==) . mapBoth pp_ <$> target_outputs `zip` prediction_outputs
                                    in and output_matches
                    return outputs_match
                -- lift . liftIO $ incProgress pb 1
                return (ok, try_num + 1)

    let jsonLinesPath = fldr <> "/eval-acc.jsonl"
    removeFile jsonLinesPath
    let acc :: Float = mean' $ fromBool <$> outcomes
    return acc
